import pandas as pd

from allotropy.allotrope.models.adm.plate_reader.benchling._2023._09.plate_reader import (
    ContainerType,
    DataSystemDocument,
    DeviceSystemDocument,
    ImageFeatureAggregateDocument,
    ImageFeatureDocumentItem,
    MeasurementAggregateDocument,
    Model,
    OpticalImagingDeviceControlAggregateDocument,
    OpticalImagingDeviceControlDocumentItem,
    OpticalImagingMeasurementDocumentItems,
    PlateReaderAggregateDocument,
    PlateReaderDocumentItem,
    ProcessedDataAggregateDocument,
    ProcessedDataDocumentItem,
    SampleDocument,
)
from allotropy.allotrope.models.shared.definitions.custom import (
    TQuantityValueMilliSecond,
    TQuantityValueNumber,
    TQuantityValueUnitless,
)
from allotropy.constants import ASM_CONVERTER_VERSION
from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.constants import NOT_APPLICABLE
from allotropy.parsers.mabtech_apex.mabtech_apex_contents import MabtechApexContents
from allotropy.parsers.mabtech_apex.mabtech_apex_structure import (
    PlateInformation,
    Well,
    WellList,
)
from allotropy.parsers.release_state import ReleaseState
from allotropy.parsers.utils.uuids import random_uuid_str
from allotropy.parsers.utils.values import quantity_or_none
from allotropy.parsers.vendor_parser import VendorParser


class MabtechApexParser(VendorParser):
    @property
    def display_name(self) -> str:
        return "Mabtech Apex"

    @property
    def release_state(self) -> ReleaseState:
        return ReleaseState.CANDIDATE_RELEASE

    def to_allotrope(self, named_file_contents: NamedFileContents) -> Model:
        raw_contents = pd.read_excel(named_file_contents.contents, sheet_name=None)
        contents = MabtechApexContents(raw_contents)
        data = PlateInformation.create(contents)
        wells = WellList.create(contents)
        return self._get_model(data, wells, named_file_contents.original_file_name)

    def _get_model(
        self, data: PlateInformation, wells: WellList, file_name: str
    ) -> Model:
        return Model(
            plate_reader_aggregate_document=PlateReaderAggregateDocument(
                device_system_document=DeviceSystemDocument(
                    device_identifier=NOT_APPLICABLE,
                    model_number=data.model_number,
                    equipment_serial_number=data.equipment_serial_number,
                ),
                data_system_document=DataSystemDocument(
                    file_name=file_name,
                    UNC_path=data.unc_path,
                    software_name="Apex",
                    software_version=data.software_version,
                    ASM_converter_name=self.get_asm_converter_name(),
                    ASM_converter_version=ASM_CONVERTER_VERSION,
                ),
                plate_reader_document=[
                    PlateReaderDocumentItem(
                        analyst=data.analyst,
                        measurement_aggregate_document=self.get_measurement_aggregate_document(
                            well
                        ),
                    )
                    for well in wells
                ],
            ),
            field_asm_manifest="http://purl.allotrope.org/manifests/plate-reader/BENCHLING/2023/09/plate-reader.manifest",
        )

    def get_measurement_aggregate_document(
        self, well: Well
    ) -> MeasurementAggregateDocument:
        return MeasurementAggregateDocument(
            container_type=ContainerType.well_plate,
            plate_well_count=TQuantityValueNumber(value=float(96)),
            measurement_time=self._get_date_time(well.measurement_time),
            measurement_document=[self.get_measurement_document_item(well)],
        )

    def get_measurement_document_item(
        self, well: Well
    ) -> OpticalImagingMeasurementDocumentItems:
        return OpticalImagingMeasurementDocumentItems(
            measurement_identifier=random_uuid_str(),
            sample_document=self.get_sample_document(well),
            device_control_aggregate_document=OpticalImagingDeviceControlAggregateDocument(
                device_control_document=[
                    self.get_device_control_document_item(well),
                ]
            ),
            processed_data_aggregate_document=ProcessedDataAggregateDocument(
                processed_data_document=[
                    self.get_processed_data_document(well),
                ]
            ),
        )

    def get_sample_document(self, well: Well) -> SampleDocument:
        return SampleDocument(
            sample_identifier=well.sample_identifier,
            location_identifier=well.location_identifier,
            well_plate_identifier=well.well_plate_identifier,
        )

    def get_device_control_document_item(
        self, well: Well
    ) -> OpticalImagingDeviceControlDocumentItem:
        return OpticalImagingDeviceControlDocumentItem(
            device_type="imager",
            detection_type="optical-imaging",
            exposure_duration_setting=quantity_or_none(
                TQuantityValueMilliSecond, well.exposure_duration_setting
            ),
            illumination_setting=quantity_or_none(
                TQuantityValueUnitless, well.illumination_setting
            ),
        )

    def get_processed_data_document(self, well: Well) -> ProcessedDataDocumentItem:
        return ProcessedDataDocumentItem(
            processed_data_identifier=random_uuid_str(),
            image_feature_aggregate_document=ImageFeatureAggregateDocument(
                image_feature_document=[
                    ImageFeatureDocumentItem(
                        image_feature_identifier=random_uuid_str(),
                        image_feature_name=next(iter(image_feature.keys())),
                        image_feature_result=TQuantityValueUnitless(
                            value=image_feature[next(iter(image_feature.keys()))]
                        ),
                    )
                    for image_feature in well.image_features
                ]
            ),
        )



from __future__ import annotations

from io import StringIO
import re
from typing import Any

import pandas as pd

from allotropy.named_file_contents import NamedFileContents
from allotropy.parsers.beckman_vi_cell_xr.constants import (
    DATE_HEADER,
    DEFAULT_VERSION,
    MODEL_RE,
    XrVersion,
)
from allotropy.parsers.lines_reader import read_to_lines

HEADINGS_TO_PARSER_HEADINGS = {
    "Dilution": "Dilution factor",
    "Total cells / ml (x 10^6)": "Total cells/ml (x10^6)",
    "Total cells": "Total cells",
    "Average diameter (microns)": "Avg. diam. (microns)",
    "Viable cells": "Viable cells",
    "Average circularity": "Avg. circ.",
    "Viability (%)": "Viability (%)",
    "Total viable cells / ml (x 10^6)": "Viable cells/ml (x10^6)",
    "Sample ID": "Sample ID",
    "Cell type": "Cell type",
}

NUMERIC_COLUMNS = [
    "Dilution factor",
    "Total cells/ml (x10^6)",
    "Total cells",
    "Avg. diam. (microns)",
    "Viable cells",
    "Avg. circ.",
    "Viability (%)",
    "Viable cells/ml (x10^6)",
]


class ViCellXRTXTReader:
    def __init__(self, contents: NamedFileContents) -> None:
        self.lines = read_to_lines(contents)
        self.file_info = self._get_file_info()
        self.file_version = self._get_file_version()
        self.data = self._read_data()

    def _read_data(self) -> pd.DataFrame:
        # generate long, single column dataframe
        file_data = pd.DataFrame(data=StringIO("\n".join(self.lines)))

        # strip whitespace
        file_data[0] = file_data[0].str.strip()

        # drop empty cells
        file_data = file_data.replace("", None).dropna(axis=0, how="all")

        # split on colon surrounded by any amount of space, limited to one split to avoid destroying datetimes
        file_data = file_data[file_data.columns[0]].str.split(
            r"\s*:\s*", n=1, expand=True
        )

        # pivot the long data of a single sample to a df that is 1 wide row
        file_data["Pivot Index"] = 1
        file_data = file_data.pivot(index="Pivot Index", columns=0)

        date_header = DATE_HEADER[self.file_version]

        # rename the columns to match the existing parser that was based on xls(x) files
        names = []
        for name in [x[1] for x in file_data.columns]:
            if name == "RunDate":
                newname = date_header
            elif name in HEADINGS_TO_PARSER_HEADINGS:
                newname = HEADINGS_TO_PARSER_HEADINGS[name]
            else:
                newname = name
            names.append(newname)

        file_data.columns = pd.Index(names)

        # convert numeric columns to numeric types
        file_data[NUMERIC_COLUMNS] = file_data[NUMERIC_COLUMNS].astype(float)

        # Do the datetime conversion and remove all rows that fail to pass as datetime
        # This fixes an issue where some files have a hidden invalid first row
        file_data[date_header] = pd.to_datetime(
            file_data[date_header].astype(str),
            format="%d %b %Y  %I:%M:%S %p",
            errors="coerce",
        )

        return file_data

    def _get_file_info(self) -> pd.Series[Any]:
        data = self.lines

        info: pd.Series[Any] = pd.Series(
            [data[0], data[3], data[8]],
            copy=False,
            index=["model", "filepath", "serial"],
        )
        return info

    def _get_file_version(self) -> XrVersion:
        match = re.match(MODEL_RE, self.file_info["model"], flags=re.IGNORECASE)
        try:
            version = match.groupdict()["version"]  # type: ignore[union-attr]
        except AttributeError:
            return DEFAULT_VERSION
        # TODO: raise exception for unsupported versions
        version = ".".join(version.split(".")[0:2])
        return XrVersion(version)
